@@ -9,13 +9,15 @@ from xml.dom import minidom
 
 import requests
 
-__version__ = "2.0.0.26"
+__version__ = "2.0.0.35"
 
 # internal module helper variables and functions
 _pma_sessions = dict()
 _pma_slideinfos = dict()
 _pma_pmacoreliteURL = "http://localhost:54001/"
 _pma_pmacoreliteSessionID = "SDK.Python"
+_pma_usecachewhenretrievingtiles = True
+_pma_amount_of_data_downloaded = {_pma_pmacoreliteSessionID: 0}
 
 def _pma_session_id(sessionID = None):
 	if (sessionID is None):
@@ -55,10 +57,13 @@ def _pma_url(sessionID = None):
 		return _pma_pmacoreliteURL
 	else:
 		# assume sessionID is a valid session; otherwise the following will generate an error
-		url = _pma_sessions[sessionID]
-		if (not url.endswith("/")):
-			url = url + "/"
-		return url
+		if sessionID in _pma_sessions.keys():
+			url = _pma_sessions[sessionID]
+			if (not url.endswith("/")):
+				url = url + "/"
+			return url
+		else:
+			raise Exception("Invalid sessionID:", sessionID)
 
 def _pma_is_lite(pmacoreURL = _pma_pmacoreliteURL):
 	url = _pma_join(pmacoreURL, "api/xml/IsLite")
@@ -85,7 +90,8 @@ def _pma_api_url(sessionID = None, xml = True):
 def _pma_join(*s):
 	joinstring = ""
 	for ss in s:
-		joinstring = os.path.join(joinstring, ss)
+		if not (ss is None):
+			joinstring = os.path.join(joinstring, ss)
 	return joinstring.replace("\\", "/")
 	
 def _pma_XmlToStringArray(root, limit = 0):
@@ -132,9 +138,12 @@ def connect(pmacoreURL = _pma_pmacoreliteURL, pmacoreUsername = "", pmacorePassw
 	Attempt to connect to PMA.core instance; success results in a SessionID
 	"""
 	if (pmacoreURL == _pma_pmacoreliteURL):
-		# no point authenticating localhost / PMA.core.lite
-		return _pma_pmacoreliteSessionID
-		
+		if is_lite():
+			# no point authenticating localhost / PMA.core.lite
+			return _pma_pmacoreliteSessionID
+		else:
+			return None
+			
 	# purposefully DON'T use helper function _pma_api_url() here:	
 	# why? Because_pma_api_url() takes session information into account (which we don't have yet)
 	url = _pma_join(pmacoreURL, "api/xml/authenticate?caller=SDK.Python") 
@@ -143,8 +152,12 @@ def connect(pmacoreURL = _pma_pmacoreliteURL, pmacoreUsername = "", pmacorePassw
 	if (pmacorePassword != ""):
 		url += "&password=" + _pma_q(pmacorePassword)
 	
-	contents = urlopen(url).read()
-	dom = minidom.parseString(contents)
+	try:
+		contents = urlopen(url).read()
+		dom = minidom.parseString(contents)
+	except:
+		# Something went wrong; unable to communicate with specified endpoint
+		return None
 		
 	loginresult = dom.firstChild
 	succ = loginresult.getElementsByTagName("Success")[0]
@@ -159,9 +172,27 @@ def connect(pmacoreURL = _pma_pmacoreliteURL, pmacoreUsername = "", pmacorePassw
 		_pma_sessions[sessionID] = pmacoreURL
 		global _pma_slideinfos
 		_pma_slideinfos[sessionID] = dict()
+		global _pma_amount_of_data_downloaded
+		_pma_amount_of_data_downloaded[sessionID] = len(contents)
 	
 	return (sessionID)	
 
+def disconnect(sessionID = None):
+	"""
+	Attempt to connect to PMA.core instance; success results in a SessionID
+	"""
+	sessionID = _pma_session_id(sessionID)
+	url = _pma_api_url(sessionID) + "DeAuthenticate?sessionID=" + _pma_q((sessionID))
+	contents = urlopen(url).read()
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(contents)
+	if (len(_pma_sessions.keys()) > 0):
+		# yes we do! This means that when there's a PMA.core active session AND PMA.core.lite version running, 
+		# the PMA.core active will be selected and returned
+		del _pma_sessions[sessionID]
+		del _pma_slideinfos[sessionID]
+	return True
+
 def get_root_directories(sessionID = None):
 	"""
 	Return an array of root-directories available to sessionID
@@ -169,6 +200,8 @@ def get_root_directories(sessionID = None):
 	sessionID = _pma_session_id(sessionID)
 	url = _pma_api_url(sessionID) + "GetRootDirectories?sessionID=" + _pma_q((sessionID))
 	contents = urlopen(url).read()
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(contents)
 	dom = minidom.parseString(contents)
 	return _pma_XmlToStringArray(dom.firstChild)
 
@@ -177,28 +210,36 @@ def get_directories(startDir, sessionID = None):
 	Return an array of sub-directories available to sessionID in the startDir directory
 	"""
 	sessionID = _pma_session_id(sessionID)
-	url = _pma_api_url(sessionID) + "GetDirectories?sessionID=" + _pma_q(sessionID) + "&path=" + _pma_q(startDir)
-	contents = urlopen(url).read()
-	dom = minidom.parseString(contents)
-	return _pma_XmlToStringArray(dom.firstChild)
+	url = _pma_api_url(sessionID, False) + "GetDirectories?sessionID=" + _pma_q(sessionID) + "&path=" + _pma_q(startDir)
+	r = requests.get(url)
+	json = r.json()
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(json)
+	if ("Code" in json):
+		raise Exception("get_directories to " + startDir + " resulted in: " + json["Message"] + " (keep in mind that startDir is case sensitive!)")
+	elif ("d" in json):
+		dirs  = json["d"]
+	else:
+		dirs = json
+	return dirs
 
 def get_first_non_empty_directory(startDir = None, sessionID = None):
 	sessionID = _pma_session_id(sessionID)
 
 	if ((startDir is None) or (startDir == "")):
 		startDir = "/"
-	slides = get_slides(startDir, sessionID)
+	slides = get_slides(startDir = startDir, sessionID = sessionID)
 	if (len(slides) > 0):
 		return startDir
 	else:
 		if (startDir == "/"):
-			for dir in get_root_directories(sessionID):
-				nonEmtptyDir = get_first_non_empty_directory(dir, sessionID)
+			for dir in get_root_directories(sessionID = sessionID):
+				nonEmtptyDir = get_first_non_empty_directory(startDir = dir, sessionID = sessionID)
 				if (not (nonEmtptyDir is None)):
 					return nonEmtptyDir
 		else:
 			for dir in get_directories(startDir, sessionID):
-				nonEmtptyDir = get_first_non_empty_directory(dir, sessionID)
+				nonEmtptyDir = get_first_non_empty_directory(startDir = dir, sessionID = sessionID)
 				if (not (nonEmtptyDir is None)):
 					return nonEmtptyDir
 	return None
@@ -210,18 +251,26 @@ def get_slides(startDir, sessionID = None):
 	sessionID = _pma_session_id(sessionID)
 	if (startDir.startswith("/")):
 		startDir = startDir[1:]		
-	url = _pma_api_url(sessionID) + "GetFiles?sessionID=" + _pma_q(sessionID) + "&path=" + _pma_q(startDir)	
-	contents = urlopen(url).read()
-	dom = minidom.parseString(contents)
-	return _pma_XmlToStringArray(dom.firstChild)
+	url = _pma_api_url(sessionID, False) + "GetFiles?sessionID=" + _pma_q(sessionID) + "&path=" + _pma_q(startDir)	
+	r = requests.get(url)
+	json = r.json()
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(json)
+	if ("Code" in json):
+		raise Exception("get_slides from " + startDir + " resulted in: " + json["Message"] + " (keep in mind that startDir is case sensitive!)")
+	elif ("d" in json):
+		slides  = json["d"]
+	else:
+		slides = json
+	return slides
 
-def get_slide_file_extension(slideRef, sessionID = None):
+def get_slide_file_extension(slideRef):
 	"""
 	Determine the file extension for this slide
 	"""
 	return os.path.splitext(slideRef)[-1]
 
-def get_slide_file_name(slideRef, sessionID = None):
+def get_slide_file_name(slideRef):
 	"""
 	Determine the file name (with extension) for this slide
 	"""
@@ -234,6 +283,8 @@ def get_uid(slideRef, sessionID = None):
 	sessionID = _pma_session_id(sessionID)
 	url = _pma_api_url(sessionID) + "GetUID?sessionID=" + _pma_q(sessionID) + "&path=" + _pma_q(slideRef)
 	contents = urlopen(url).read()
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(contents)
 	dom = minidom.parseString(contents)
 	return _pma_XmlToStringArray(dom)[0]
 	
@@ -272,8 +323,11 @@ def get_slide_info(slideRef, sessionID = None):
 
 	if (not (slideRef in _pma_slideinfos[sessionID])):
 		url = _pma_api_url(sessionID, False) + "GetImageInfo?SessionID=" + _pma_q(sessionID) +  "&pathOrUid=" + _pma_q(slideRef)
+		print(url);
 		r = requests.get(url)
 		json = r.json()
+		global _pma_amount_of_data_downloaded 
+		_pma_amount_of_data_downloaded[sessionID] += len(json)
 		if ("Code" in json):
 			raise Exception("ImageInfo to " + slideRef + " resulted in: " + json["Message"] + " (keep in mind that slideRef is case sensitive!)")
 		elif ("d" in json):
@@ -288,10 +342,22 @@ def get_max_zoomlevel(slideRef, sessionID = None):
 	Determine the maximum zoomlevel that still represents an optical magnification
 	"""
 	info = get_slide_info(slideRef, sessionID)
-	if ("MaxZoomLevel" in info): 
-		return int(info["MaxZoomLevel"])
+	if (info == None):
+		print("Unable to get information for", slideRef, " from ", sessionID)
+		return 0
 	else:
-		return int(info["NumberOfZoomLevels"])
+		if ("MaxZoomLevel" in info): 
+			try:
+				return int(info["MaxZoomLevel"])
+			except:
+				print("Something went wrong consulting the MaxZoomLevel key in info{} dictionary; value =", info["MaxZoomLevel"])
+				return 0
+		else:
+			try:
+				return int(info["NumberOfZoomLevels"])
+			except:
+				print("Something went wrong consulting the NumberOfZoomLevels key in info{} dictionary; value =", info["NumberOfZoomLevels"])
+				return 0
 
 def get_zoomlevels_list(slideRef, sessionID = None, min_number_of_tiles = 0):
 	"""
@@ -309,8 +375,8 @@ def get_zoomlevels_dict(slideRef, sessionID = None, min_number_of_tiles = 0):
 		n = total number of tiles at specified zoomlevel (x * y)
 	Use min_number_of_tiles argument to specify that you're only interested in zoomlevels that include at lease a given number of tiles
 	"""
-	zoomlevels = list(range(0, get_max_zoomlevel(slideRef) + 1))
-	dimensions = [ get_number_of_tiles(slideRef, z) for z in zoomlevels if get_number_of_tiles(slideRef, z)[2] > min_number_of_tiles]
+	zoomlevels = list(range(0, get_max_zoomlevel(slideRef, sessionID) + 1))
+	dimensions = [ get_number_of_tiles(slideRef, z, sessionID) for z in zoomlevels if get_number_of_tiles(slideRef, z, sessionID)[2] > min_number_of_tiles]
 	d = dict(zip(zoomlevels[-len(dimensions):], dimensions))
 	
 	
@@ -405,6 +471,8 @@ def get_barcode_image(slideRef, sessionID = None):
 	"""Get the barcode (alias for "label") image for a slide"""
 	r = requests.get(get_barcode_url(slideRef, sessionID))
 	img = Image.open(BytesIO(r.content))
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(r.content)
 	return img
 
 def get_label_url(slideRef, sessionID = None):
@@ -415,6 +483,8 @@ def get_label_image(slideRef, sessionID = None):
 	"""Get the label image for a slide"""
 	r = requests.get(get_label_url(slideRef, sessionID))
 	img = Image.open(BytesIO(r.content))
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(r.content)
 	return img
 		
 def get_thumbnail_url(slideRef, sessionID = None):
@@ -429,14 +499,25 @@ def get_thumbnail_image(slideRef, sessionID = None):
 	"""Get the thumbnail image for a slide"""
 	r = requests.get(get_thumbnail_url(slideRef, sessionID))
 	img = Image.open(BytesIO(r.content))
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(r.content)
 	return img
 
-def get_tile(slideRef, x = 0, y = 0, zoomlevel = None, sessionID = None):
-	"""Get a single tile at position (x, y)"""
+def get_tile(slideRef, x = 0, y = 0, zoomlevel = None, sessionID = None, format = "jpg", quality = 100): 
+	"""
+	Get a single tile at position (x, y)
+	Format can be 'jpg' or 'png'
+	Quality is an integer value and varies from 0 (as much compression as possible; not recommended) to 100 (100%, no compression)
+	"""
 	sessionID = _pma_session_id(sessionID)
 	if (zoomlevel is None):
 		zoomlevel = 0   # get_max_zoomlevel(slideRef, sessionID)
-	url = (_pma_url(sessionID) + "tile"
+
+	url = _pma_url(sessionID)
+	if url is None:
+		raise Exception("Unable to determine the PMA.core instance belonging to " + sessionID)
+
+	url += ("tile"
 		+ "?SessionID=" + _pma_q(sessionID)
 		+ "&channels=" + _pma_q("0")
 		+ "&timeframe=" + _pma_q("0")
@@ -444,14 +525,24 @@ def get_tile(slideRef, x = 0, y = 0, zoomlevel = None, sessionID = None):
 		+ "&pathOrUid=" + _pma_q(slideRef)
 		+ "&x=" + _pma_q(x)
 		+ "&y=" + _pma_q(y)
-		+ "&z=" + _pma_q(zoomlevel)	)
+		+ "&z=" + _pma_q(zoomlevel)	
+		+ "&format=" + _pma_q(format)
+		+ "&quality=" + _pma_q(quality)
+		+ "&cache=" + str(_pma_usecachewhenretrievingtiles).lower())
 	r = requests.get(url)
 	img = Image.open(BytesIO(r.content))
+	global _pma_amount_of_data_downloaded 
+	_pma_amount_of_data_downloaded[sessionID] += len(r.content)
 	return img
 
-def get_tiles(slideRef, fromX = 0, fromY = 0, toX = None, toY = None, zoomlevel = None, sessionID = None):
-	"""Get all tiles with a (fromX, fromY, toX, toY) rectangle. Navigate left to right, top to bottom"""
+def get_tiles(slideRef, fromX = 0, fromY = 0, toX = None, toY = None, zoomlevel = None, sessionID = None, format = "jpg", quality = 100):
+	"""
+	Get all tiles with a (fromX, fromY, toX, toY) rectangle. Navigate left to right, top to bottom
+	Format can be 'jpg' or 'png'
+	Quality is an integer value and varies from 0 (as much compression as possible; not recommended) to 100 (100%, no compression)
+	"""
 	sessionID = _pma_session_id(sessionID)
+
 	if (zoomlevel is None):
 		zoomlevel = 0   # get_max_zoomlevel(slideRef, sessionID)
 	if (toX is None):
@@ -460,7 +551,7 @@ def get_tiles(slideRef, fromX = 0, fromY = 0, toX = None, toY = None, zoomlevel
 		toY = get_number_of_tiles(slideRef, zoomlevel, sessionID)[1]
 	for x in range(fromX, toX):
 		for y in range(fromY, toY):
-			yield get_tile(slideRef, x, y, zoomlevel, sessionID)
+			yield get_tile(slideRef = slideRef, x = x, y = y, zoomlevel = zoomlevel, sessionID = sessionID, format = format, quality = quality)
 			
 def show_slide(slideRef, sessionID = None):
 	"""Launch the default webbrowser and load a web-based viewer for the slide"""
@@ -473,7 +564,12 @@ def show_slide(slideRef, sessionID = None):
 	if (sessionID == _pma_pmacoreliteSessionID):
 		url = "http://free.pathomation.com/pma-view-lite/?path=" + _pma_q(slideRef)
 	else:
-		url = (_pma_url(sessionID) + "viewer/index.htm"
+		url = _pma_url(sessionID)
+		if url is None:
+			raise Exception("Unable to determine the PMA.core instance belonging to " + sessionID)
+		else:
+			url += ("viewer/index.htm"
 			+ "?sessionID=" + _pma_q(sessionID)
 			+ "^&pathOrUid=" + _pma_q(slideRef))    # note the ^& to escape a regular &
 	os.system(os_cmd+url)
+6
\ No newline at end of file
